from celery_longterm_scheduler import get_scheduler
from celery_longterm_scheduler.conftest import CELERY
from unittest import mock
import pendulum


@CELERY.task
def echo(arg):
    return arg


def test_should_store_all_arguments_needed_for_send_task(celery_worker):
    # Cannot do this with a Mock, since they (technically correctly)
    # differentiate recording calls between args and kw, so a call
    # `send_task(1, 2,  3)` is not considered equal to
    # `send_task(1, args=2, kwargs=3)`, although semantically it is the same.
    def record_task(
            name, args=None, kwargs=None, countdown=None, eta=None,
            task_id=None, producer=None, connection=None, router=None,
            result_cls=None, expires=None, publisher=None, link=None,
            link_error=None, add_to_parent=True, group_id=None, retries=0,
            chord=None, reply_to=None, time_limit=None, soft_time_limit=None,
            root_id=None, parent_id=None, route_name=None, shadow=None,
            chain=None, task_type=None, **options):
        options.update(dict(
            args=args, kwargs=kwargs, countdown=countdown,
            eta=eta, task_id=task_id, producer=producer, connection=connection,
            router=router, result_cls=result_cls, expires=expires,
            publisher=publisher, link=link, link_error=link_error,
            add_to_parent=add_to_parent, group_id=group_id, retries=retries,
            chord=chord, reply_to=reply_to, time_limit=time_limit,
            soft_time_limit=soft_time_limit, root_id=root_id,
            parent_id=parent_id, route_name=route_name, shadow=shadow,
            chain=chain, task_type=task_type
        ))
        calls.append((name, options))
    calls = []

    with mock.patch.object(CELERY, 'send_task', new=record_task):
        result = echo.apply_async(('foo',), eta=pendulum.now())
        task = get_scheduler(CELERY).backend.get(result.id)
        args = task[0]
        kw = task[1]
        # schedule() always generates an ID itself (to reuse it for the
        # scheduler storage), while the normal apply_async() defers that to
        # send_task(). We undo this here for comparison purposes.
        kw['task_id'] = None
        CELERY.send_task(*args, **kw)
        scheduled_call = calls[0]

        echo.apply_async(('foo',))
        normal_call = calls[1]
        # Special edge case, see Task._schedule() for an explanation
        normal_call[1]['result_cls'] = None
        assert scheduled_call == normal_call


def test_should_bypass_if_no_eta_given():
    with mock.patch(
            'celery_longterm_scheduler.task.Task._schedule') as schedule:
        result = echo.apply_async(('foo',))
        assert schedule.call_count == 0
        result.get()  # Be careful about test isolation

        result = echo.apply_async(('foo',), eta=None)
        assert schedule.call_count == 0
        result.get()  # Be careful about test isolation

        echo.apply_async(('foo',), eta=pendulum.now())
        assert schedule.call_count == 1



from celery_longterm_scheduler import backend
import click
import celery.bin.base
import contextlib
import fcntl
import logging
import os
import pendulum


log = logging.getLogger(__name__)


class Scheduler:
    """Main scheduler functionality:

    :store: schedule tasks for later
    :revoke: revoke scheduled tasks
    :execute_pending: execute scheduled tasks due by a given timestamp

    Clients should use ``get_scheduler(app)`` with their celery app instance
    to get hold of the corresponding Scheduler instance.
    """

    CONF_KEY = '__longterm_scheduler_backend'

    def __init__(self, app):
        self.app = app
        # Singleton behaviour
        if self.CONF_KEY not in app.conf:
            app.conf[self.CONF_KEY] = backend.by_url(
                app.conf['longterm_scheduler_backend'], app)
        self.backend = app.conf[self.CONF_KEY]

    @classmethod
    def from_app(cls, app):
        return cls(app)

    def store(self, timestamp, task_id, args, kw):
        """Schedules the task (represented by the ``args`` and ``kw`` of the
        postponed send_task() call) under ``task_id`` and ``timestamp``.

        :param timestamp: timezone-aware datetime
        :param task_id: string, the task id, can be used in revoke()
        :param args: tuple, positional arguments for the task
        :param kw: dict, keyword arguments for the task
        """
        self.backend.set(timestamp, task_id, args, kw)

    def execute_pending(self, timestamp):
        """Looks up scheduled tasks that are due on or before ``timestamp``,
        creates normal celery tasks for them, and removes them from the
        scheduler storage.

        :param timestamp: timezone-aware datetime
        """
        log.info('Start executing tasks older than %s', timestamp)
        for id, task in self.backend.get_older_than(timestamp):
            self._execute_task(id, task[0], task[1])
        log.info('End executing tasks older than %s', timestamp)

    def _execute_task(self, task_id, args, kw):
        log.info('Enqueuing %s', task_id)
        # XXX No transactions, so we accept the risk of executing a task twice,
        # rather than not executing it at all (with regards to revoke failing).
        self.app.send_task(*args, **kw)
        self.revoke(task_id)

    def revoke(self, task_id):
        """Removes the task scheduled by ``store(task_id)`` from scheduler
        storage.

        :returns: True if ``task_id`` was found and removed, False otherwise"""
        try:
            self.backend.delete(task_id)
            log.info('Revoked %s', task_id)
            return True
        except KeyError:
            return False


get_scheduler = Scheduler.from_app


@click.command(name='longterm_scheduler', cls=celery.bin.base.CeleryCommand)
@click.option(
    '--timestamp', default='now',
    help='Execute tasks older/equal to TIMESTAMP, default: now')
@click.option(
    '--lockfile', default='',
    help='Path to lockfile, to prevent multiple simultaneous runs')
@click.pass_context
def main(ctx, timestamp, lockfile):
    """The subcommand ``celery longterm_scheduler`` executes scheduled tasks
    that are due on or before a given time (default: now), by creating normal
    celery tasks for them.
    """
    app = ctx.obj.app
    app.log.setup(
        logging.WARNING if ctx.parent.params.get('quiet') else logging.INFO)
    # The `tz` parameter applies only if no timezone information is
    # present in the string -- which is precisely what we want here;
    # tz=None means use the locale's timezone.
    timestamp = pendulum.parse(timestamp, tz=None)
    if lockfile:
        with locked(lockfile):
            get_scheduler(app).execute_pending(timestamp)
    else:
        get_scheduler(app).execute_pending(timestamp)


@contextlib.contextmanager
def locked(filename):
    """Context manager that acquires a file-based lock or raises RuntimError if
    already locked.
    """
    lockfile = open(filename, 'a+')
    try:
        fcntl.lockf(lockfile, fcntl.LOCK_EX | fcntl.LOCK_NB)
    except IOError:
        raise RuntimeError('Another process is already running')
    # Publishing the process id is handy for debugging.
    lockfile.seek(0)
    lockfile.truncate()
    lockfile.write('%s\n' % os.getpid())
    lockfile.flush()
    try:
        yield
    finally:
        lockfile.seek(0)
        lockfile.truncate()
        lockfile.close()  # This implicitly unlocks.



import base64
import binascii
import celery.backends.redis
import collections
import json
import pendulum
import pickle
import redis


class AbstractBackend:
    """Interface for the scheduler storage backend. Also see test_backend.py
    for the corresponding contract tests that every implementation must pass.

    Clients should not instantiate a backend class themselves, but rather
    use the ``by_url()`` mechanism.
    """

    def __init__(self, url, app):
        """
        :param url: string, a backend-specific configuration address
        :param app: celery App instance (mainly so the backend can access the
        configuration in ``app.conf``)
        """
        raise NotImplementedError()

    def set(self, timestamp, task_id, args, kw):
        """Stores ``args`` and ``kw`` under ``task_id`` and ``timestamp``.
        args and kw are serialized using JSON.

        :param timestamp: timezone-aware datetime
        :param task_id: string
        :param args: tuple, positional arguments for the task
        :param kw: dict, keyword arguments for the task
        """
        raise NotImplementedError()

    def get(self, task_id):
        """Retrieves a task entry stored by ``set()``

        :param task_id: string
        :returns: tuple (args, kw) -- args tuple, kw dict
        """
        raise NotImplementedError()

    def delete(self, task_id):
        """Removes a task entry stored by ``set()``.

        :raises: KeyError if ``task_id`` was not found
        """
        raise NotImplementedError()

    def get_older_than(self, timestamp):
        """Retrieves task entries scheduled for times older or equal than
        ``timestamp``.

        :param timestamp: timezone-aware datetime
        :return: iterable of tuple (task_id, (args, kw))
        """
        raise NotImplementedError()


class MemoryBackend(AbstractBackend):
    """In-memory backend implementation, for tests."""

    def __init__(self, unused_url, unused_app):
        self.by_id = {}
        self.by_time = collections.defaultdict(list)

    def set(self, timestamp, task_id, args, kw):
        if timestamp.tzinfo is None:
            raise ValueError('Timezone required, got %s', timestamp)
        self.by_id[task_id] = serialize([args, kw])
        self.by_time[serialize_timestamp(timestamp)].append(task_id)

    def get(self, task_id):
        args, kw = deserialize(self.by_id[task_id])
        if isinstance(kw.get('args'), list):
            kw['args'] = tuple(kw['args'])
        return (tuple(args), kw)

    def delete(self, task_id):
        del self.by_id[task_id]
        for ts, id in self.by_time.items():
            if id == task_id:
                break
        bucket = self.by_time[ts]
        bucket.remove(task_id)
        if not bucket:
            del self.by_time[ts]

    def get_older_than(self, timestamp):
        timestamp = serialize_timestamp(timestamp)
        for ts in sorted(self.by_time.keys()):
            if ts > timestamp:
                break
            for id in self.by_time[ts]:
                yield (id, self.get(id))


class RedisBackend(AbstractBackend):
    """Default backend implementation: redis"""

    redis = redis
    # This is persisted in redis, only change when also having a migration plan
    BY_TIME_KEY = 'scheduled_task_id_by_time'

    def __init__(self, url, app):
        self.url = url
        self.app = app
        # Taken from celery.backends.redis.RedisBackend.__init__()
        max_connections = app.conf.get('redis_max_connections')
        socket_timeout = app.conf.get('redis_socket_timeout')
        socket_connect_timeout = app.conf.get('redis_socket_connect_timeout')
        self.connparams = {
            'max_connections': max_connections,
            'socket_timeout': socket_timeout and float(socket_timeout),
            'socket_connect_timeout':
                socket_connect_timeout and float(socket_connect_timeout),
        }
        # Sneaky "inherit this one method only" transplant
        self._params_from_url = celery.backends.redis.RedisBackend.\
            _params_from_url.__get__(self)
        self.connparams = self._params_from_url(url, self.connparams)
        # We probably don't need a parameterizeable ConnectionPool, like
        # celery.backends.redis.RedisBackend._create_client() does.
        self.client = self.redis.StrictRedis(
            connection_pool=self.redis.ConnectionPool(**self.connparams))

    def set(self, timestamp, task_id, args, kw):
        if timestamp.tzinfo is None:
            raise ValueError('Timezone required, got %s', timestamp)
        timestamp = serialize_timestamp(timestamp)
        self.client.set(task_id, serialize([args, kw]))
        self.client.zadd(self.BY_TIME_KEY, mapping={task_id: timestamp})

    def get(self, task_id):
        task = self.client.get(task_id)
        if task is None:
            raise KeyError(task_id)
        args, kw = deserialize(task)
        if isinstance(kw.get('args'), list):
            kw['args'] = tuple(kw['args'])
        return (tuple(args), kw)

    def delete(self, task_id):
        removed = 0
        removed += self.client.delete(task_id)
        removed += self.client.zrem(self.BY_TIME_KEY, task_id)
        if removed != 2:
            raise KeyError(task_id)

    def get_older_than(self, timestamp):
        timestamp = serialize_timestamp(timestamp)
        for id in self.client.zrangebyscore(self.BY_TIME_KEY, 0, timestamp):
            # Typically celery uses uuid, so ascii would suffice, but who knows
            # what kind of ids random applications use in the wild.
            yield (id.decode('utf-8'), self.get(id))


# Could be made extensible via entrypoints, like in celery.app.backends.
BACKENDS = {
    'memory': MemoryBackend,
    'redis': RedisBackend,
    'rediss': RedisBackend,
}


def by_url(url, app):
    if '://' not in url:
        raise ValueError(
            'longterm_scheduler_backend must be an URL, got %r' % url)
    scheme = url.split('://')[0]
    return BACKENDS[scheme](url, app)


class PickleFallbackJSONEncoder(json.JSONEncoder):
    """Serializes non-native JSON types using pickle.

    We need this mostly because Task.apply_async() needs to store itself (the
    Task instance), since send_task() needs it e.g. for routing information.
    So we *hope* that nobody puts anything non-pickleable onto their tasks.
    """

    PICKLE_MARKER = '__python_pickle__'

    def default(self, o):
        raw = pickle.dumps(o)
        raw = base64.b64encode(raw).decode('ascii')
        return self.PICKLE_MARKER + raw

    @classmethod
    def decode_dict(cls, o):
        for key, value in o.items():
            if isinstance(value, str) and value.startswith(cls.PICKLE_MARKER):
                raw = value.replace(cls.PICKLE_MARKER, '', 1)
                try:
                    raw = base64.b64decode(raw)
                except binascii.Error:
                    # We hopefully have a py2 pickle.
                    raw = raw.encode('ascii')
                o[key] = pickle.loads(raw)
        return o


def serialize(obj):
    return json.dumps(obj, cls=PickleFallbackJSONEncoder)


def deserialize(string):
    return json.loads(
        string, object_hook=PickleFallbackJSONEncoder.decode_dict)


def serialize_timestamp(timestamp):
    """Converts a datetime into seconds since the epoch."""
    return int(pendulum.instance(timestamp).timestamp())



from celery_longterm_scheduler.conftest import CELERY
from unittest import mock
import celery_longterm_scheduler
import pendulum
import time


PAST_DATE = pendulum.datetime(2017, 1, 20)
FUTURE_DATE = pendulum.datetime(3000, 1, 20)
record_calls = []


@CELERY.task
def record(arg):
    record_calls.append(arg)


def test_execute_pending_does_not_run_tasks_scheduled_in_the_future(
        celery_worker):
    record_calls[:] == []
    due = FUTURE_DATE
    record.apply_async(('not_yet',), eta=due)
    assert not record_calls
    scheduler = celery_longterm_scheduler.get_scheduler(CELERY)
    # Try to execute task, scheduled in the future
    scheduler.execute_pending(pendulum.now())
    # XXX I don't think there is any "job completed" signal we could wait for.
    time.sleep(1)
    assert not record_calls


def test_execute_pending_runs_scheduled_tasks(celery_worker):
    record_calls[:] == []
    due = PAST_DATE
    record.apply_async(('foo',), eta=due)
    assert not record_calls
    scheduler = celery_longterm_scheduler.get_scheduler(CELERY)
    # Try to execute task, scheduled in the past
    scheduler.execute_pending(pendulum.now())
    # XXX I don't think there is any "job completed" signal we could wait for.
    time.sleep(1)
    assert record_calls == ['foo']


@CELERY.task
def echo(arg):
    return arg


def test_execute_pending_deletes_from_storage():
    due = PAST_DATE
    echo.apply_async(('foo',), eta=due)

    scheduler = celery_longterm_scheduler.get_scheduler(CELERY)
    pending = list(scheduler.backend.get_older_than(due))
    assert pending

    with mock.patch.object(CELERY, 'send_task'):
        scheduler.execute_pending(due)
    pending = list(scheduler.backend.get_older_than(due))
    assert not pending


def test_revoke_deletes_from_storage():
    due = PAST_DATE
    id = echo.apply_async(('foo',), eta=due).id
    scheduler = celery_longterm_scheduler.get_scheduler(CELERY)
    pending = list(scheduler.backend.get_older_than(due))
    assert pending

    assert scheduler.revoke(id)
    pending = list(scheduler.backend.get_older_than(due))
    assert not pending


def test_revoke_returns_false_for_nonexistent_id():
    scheduler = celery_longterm_scheduler.get_scheduler(CELERY)
    assert scheduler.revoke('nonexistent') is False



import celery
import celery.utils
import celery_longterm_scheduler


class Task(celery.Task):
    """Task subclass that adds special behaviour when ``apply_async()`` is
    called with an ``eta`` argument: the job is not created immediately, but
    rather stored in a separate scheduler storage.

    This still returns a normal ``AsyncResult`` object, but only getting the
    ``id`` is supported on it. This ID can be passed to
    ``celery_longterm_scheduler.revoke()`` to remove the scheduled job from
    storage.
    """

    def apply_async(self, args=None, kwargs=None, task_id=None, producer=None,
                    link=None, link_error=None, shadow=None, **options):
        if options.get('eta') is not None:
            timestamp = options.pop('eta')

            # copy&paste from celery.app.task.Task.apply_async()
            if self.__v2_compat__:
                shadow = shadow or self.shadow_name(
                    self(), args, kwargs, options)
            else:
                shadow = shadow or self.shadow_name(args, kwargs, options)

            preopts = self._get_exec_options()
            options = dict(preopts, **options) if options else preopts

            options.setdefault('ignore_result', self.ignore_result)
            if self.priority:
                options.setdefault('priority', self.priority)
            # end copy&paste

            return self._schedule(
                timestamp,
                args=args, kwargs=kwargs, task_id=task_id, producer=producer,
                link=link, link_error=link_error, shadow=shadow, **options)
        else:
            return super(Task, self).apply_async(
                args=args, kwargs=kwargs, task_id=task_id, producer=producer,
                link=link, link_error=link_error, shadow=shadow, **options)

    def _schedule(self, timestamp, **kw):
        # Store parameters apply_async() passes to app.send_task() in addition
        # to its own **kw.
        kw['task_type'] = self
        # We don't set result_cls, since serializing instancemethods is a pain
        # and the additional settings of self.AsyncResult compared to
        # app.AsyncResult don't make a difference _inside_ send_task, so we
        # don't actually need it. And for the return value of apply_async we
        # call it ourselves anyway, see below.
        # kw['result_cls'] = self.AsyncResult

        # We use the celery task_id also for our scheduler storage; this is
        # mostly for integration purposes, e.g. so that other Task subclasses
        # can be in control of the task_id and still work when inheriting us.
        if not kw.get('task_id'):
            kw['task_id'] = celery.utils.gen_unique_id()

        scheduler = celery_longterm_scheduler.get_scheduler(self.app)
        scheduler.store(timestamp, kw['task_id'], (self.name,), kw)
        return self.AsyncResult(kw['task_id'])



"""Contract test that every scheduler storage backend must comply with."""
from datetime import datetime
import celery
import celery_longterm_scheduler.backend
import celery_longterm_scheduler.conftest
import json
import pendulum
import pytest


ANYTIME = pendulum.datetime(2017, 1, 20)


@pytest.fixture(params=['memory://', 'redis://'])
def backend(request, redis_server):
    url = request.param
    if url == 'redis://':
        url += '{host}:{port}/{db}'.format(**redis_server.dsn())
    dummyapp = celery.Celery()
    return celery_longterm_scheduler.backend.by_url(url, dummyapp)


def test_task_passed_to_set_can_be_retrieved_with_get(backend):
    args = ('arg',)
    kw = {'kw': None}
    backend.set(ANYTIME, 'myid', args, kw)
    task = backend.get('myid')
    assert task[0] == args
    assert task[1] == kw


def test_set_requires_timezone_aware_datetime(backend):
    with pytest.raises(ValueError):
        backend.set(datetime.now(), 'myid', (), {})


def test_set_works_with_datetime(backend):
    due = datetime.now(pendulum.timezone('UTC'))
    backend.set(due, 'myid', (), {})
    tasks = list(backend.get_older_than(due))
    assert len(tasks) == 1
    assert tasks[0][0] == 'myid'


def test_get_nonexistent_task_id_raises_keyerror(backend):
    with pytest.raises(KeyError):
        backend.get('nonexistent')


def test_positional_args_and_the_kw_item_called_args_use_type_tuple(backend):
    # Even though we use JSON for serialization, which uses lists instead of
    # tuples, the Python convention is that positional arguments are a tuple.
    backend.set(ANYTIME, 'myid', ('arg',), {'args': ('celery-kw-arg',)})
    task = backend.get('myid')
    assert isinstance(task[0], tuple)
    assert isinstance(task[1]['args'], tuple)


def test_delete_task_can_not_be_retrieved_with_get_or_older_than(backend):
    backend.set(ANYTIME, 'one', ('arg1',), {'kw1': None})
    backend.set(ANYTIME, 'two', ('arg2',), {'kw2': None})
    backend.delete('one')
    with pytest.raises(KeyError):
        backend.get('one')
    assert backend.get('two') == (('arg2',), {'kw2': None})
    assert list(backend.get_older_than(ANYTIME)) == [
        ('two', (('arg2',), {'kw2': None}))]


def test_delete_nonexistend_task_id_raises_keyerror(backend):
    with pytest.raises(KeyError):
        backend.delete('nonexistent')


def test_get_older_than_returns_timestamps_smaller_or_equal(backend):
    backend.set(pendulum.datetime(2017, 1, 1, 9), '1', (1,), {'1': 1})
    backend.set(pendulum.datetime(2017, 1, 1, 10), '2', (2,), {'2': 2})
    backend.set(pendulum.datetime(2017, 1, 1, 11), '3', (3,), {'3': 3})
    items = list(backend.get_older_than(pendulum.datetime(2017, 1, 1, 10)))
    assert len(items) == 2
    assert items[0][0] == '1'
    assert items[0][1] == ((1,), {'1': 1})
    assert items[1][0] == '2'
    assert items[1][1] == ((2,), {'2': 2})


def test_py3_deserializes_py2_pickle():
    from celery_longterm_scheduler.backend import PickleFallbackJSONEncoder
    marker = PickleFallbackJSONEncoder.PICKLE_MARKER
    py2_pickle = ("ccelery.app.registry\n_unpickle_task_v2\np0\n"
                  "(S'celery.ping'\np1\nS'celery_longterm_scheduler.conftest'"
                  "\np2\ntp3\nRp4\n.")
    loaded = celery_longterm_scheduler.backend.deserialize(
        json.dumps({'task_type': marker + py2_pickle}))['task_type']
    ping = celery_longterm_scheduler.conftest.celery_ping.__maybe_evaluate__()
    # Sigh, this proxy business prevents doing a simple object identity check.
    assert repr(type(loaded)) == repr(type(ping))



from setuptools import setup, find_packages


setup(
    name='celery_longterm_scheduler',
    version='1.3.0',
    author='Zeit Online',
    author_email='zon-backend@zeit.de',
    url='https://github.com/zeitonline/celery_longterm_scheduler',
    description="Schedules celery tasks to run in the potentially far future",
    long_description='\n\n'.join(
        open(x).read() for x in ['README.rst', 'CHANGES.txt']),
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    zip_safe=False,
    license='BSD',
    install_requires=[
        'celery>=5.0.0.dev0',
        'click',
        'pendulum',
        'redis>=3.0',
        'setuptools',
    ],
    extras_require={'test': [
        'pytest',
        'testing.redis',
    ]},
    entry_points={
        'celery.commands': [
            'longterm_scheduler = celery_longterm_scheduler.scheduler:main',
        ]
    },
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],

)



import celery
import celery.contrib.testing.app
import celery.contrib.testing.worker
import celery_longterm_scheduler
import pytest
import redis
import testing.redis


CELERY = celery.Celery(task_cls=celery_longterm_scheduler.Task)


@pytest.fixture(scope='session')
def celery_worker(request):
    CELERY.conf.update(celery.contrib.testing.app.DEFAULT_TEST_CONFIG)
    CELERY.conf['longterm_scheduler_backend'] = 'memory://'
    worker = celery.contrib.testing.worker.start_worker(CELERY)
    worker.__enter__()
    request.addfinalizer(lambda: worker.__exit__(None, None, None))


# celery.contrib.testing.worker expects a 'ping' task, so it can check that the
# worker is running properly.
@CELERY.task(name='celery.ping')
def celery_ping():
    return 'pong'


@pytest.fixture(scope='session')
def redis_server_session():
    server = testing.redis.RedisServer()
    yield server
    server.stop()


@pytest.fixture
def redis_server(redis_server_session):
    yield redis_server_session
    client = redis.StrictRedis(**redis_server_session.dsn())
    for key in client.keys():
        client.delete(key)
